{
  Autor: Michael Springwald

  Datum: Dienstag, 05.05.2015
}

unit uplTCP_Synapse;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, contnrs,
  synsock,blcksock
  ;

type
  TPLTCP_Client = class;
  TPLTCP_Server = class;

  TPLTCP_OnReseived = procedure (Client:TPLTCP_Client; const msg:string) of object;

  TPLTCP_AddNewClient = procedure (const NewClient:TPLTCP_Client) of object;

  { TPLTCP_Client }
  TPLTCP_Client = class(TThread)
  private
    fFromHost: String;
    fID: string;
    fIndex: integer;
    fOnReceived: TPLTCP_OnReseived;
    LastMessage:string;
  protected
    procedure DoResivedMsg;

    procedure Execute; override;
  public
    Socket: TTCPBlockSocket;
    constructor Create(CreateSuspended: Boolean; const StackSize: SizeUInt=DefaultStackSize);
    destructor Destroy; override;

    procedure Send(aMSG:String);
    procedure StartSocket(IP: String; Port: Integer);
  published
    property OnReceived:TPLTCP_OnReseived read fOnReceived write fOnReceived;
    property FromHost:String read fFromHost write fFromHost;
    property Index:integer read fIndex write fIndex;
    property ID:string read fID write fID;
  end; // TPLTCP_Client

  { TPLTCP_ClientList }
  TPLTCP_ClientList = class
  private
    fOnReceived: TPLTCP_OnReseived;
    function GetCount: Integer;
    function GetItem(index: integer): TPLTCP_Client;
  protected
    procedure Reseived(Client:TPLTCP_Client; const msg:string);
  public
    Items:TObjectLIst;
    Server:TPLTCP_Server;
    constructor Create;
    destructor Destroy; override;

    function AddClient(const FromHost, aID:String):TPLTCP_Client;
    property Item[index:integer]:TPLTCP_Client read GetItem; default;
    property Count:Integer read GetCount;

  published
    property OnReceived:TPLTCP_OnReseived read fOnReceived write fOnReceived;
  end; // TPLTCP_ClientList

  { TPLTCP_Server }
  TPLTCP_Server = class(TThread)
  private
    fHost: String;
    fOnAddClient: TPLTCP_AddNewClient;
    fOnReseived: TPLTCP_OnReseived;
    fPort: Integer;
  protected
    procedure Execute; override;
  public
    Socket: TTCPBlockSocket;
    ClientList:TPLTCP_ClientList;

    constructor Create(CreateSuspended: Boolean; const StackSize: SizeUInt=DefaultStackSize);
    destructor Destroy; override;

    procedure StartServer(const aHost:string; const aPort:Integer);
    procedure Stop;

  published
    property Port:Integer read fPort write fPort;
    property Host:String read fHost write fHost;
    property OnAddClient:TPLTCP_AddNewClient read fOnAddClient write fOnAddClient;
    property OnReseived:TPLTCP_OnReseived read fOnReseived write fOnReseived;
  end; // TPLTCP_Server

implementation

{ TPLTCP_ClientList }

function TPLTCP_ClientList.GetCount: Integer;
begin
  result:=Items.Count;
end; // TPLTCP_ClientList.GetCount

function TPLTCP_ClientList.GetItem(index: integer): TPLTCP_Client;
begin
  result:=Items[index] as TPLTCP_Client;
end; // TPLTCP_ClientList.GetItem

procedure TPLTCP_ClientList.Reseived(Client: TPLTCP_Client; const msg: string);
begin
  if Assigned(OnReceived) then begin
    if pos('LogOut',msg) > 0 then begin
      writeln('Client wird aus Liste entfernt!!!');
      Items.Remove(Client);
    end;

    OnReceived(Client,msg);
  end;
end; // TPLTCP_ClientList.Reseived

constructor TPLTCP_ClientList.Create;
begin
  inherited Create;
  Items:=TObjectList.create;
  Server:=nil;
end; // TPLTCP_ClientList.Create

destructor TPLTCP_ClientList.Destroy;
begin
  Items.Free;
  inherited Destroy;
end; // TPLTCP_ClientList.Destroy

function TPLTCP_ClientList.AddClient(const FromHost, aID: String): TPLTCP_Client;
begin
  Result:=Item[Items.Add(TPLTCP_Client.Create(true))];
  Result.FromHost:=FromHost;
  Result.Index:=Count;
  Result.ID:=aID;
  Result.OnReceived:=@Reseived;
end; // TPLTCP_ClientList.AddClient

procedure TPLTCP_Client.DoResivedMsg;
var
  x:integer;
  str:String;
begin
  x:=Pos('NewClient:',LastMessage);
  if x >= 0 then begin
    str:=copy(LastMessage,x+10,Length(LastMessage));
    ID:=str;
  end;

  if Assigned(fOnReceived) then
    fOnReceived(self,LastMessage);

  LastMessage:='';
end; // TPLTCP_Client.DoResivedMsg

{ TPLTCP_Client }
constructor TPLTCP_Client.Create(CreateSuspended: Boolean; const StackSize: SizeUInt=DefaultStackSize);
begin
  inherited Create(CreateSuspended,StackSize);
  fOnReceived:=nil;
  fIndex:=-1;
  fFromHost:='';
  LastMessage:='';

  fID:='';
end; // TPLTCP_Client.Create

destructor TPLTCP_Client.Destroy;
begin
  Send('LogOut:'+ID);
  inherited Destroy;
end; // TPLTCP_Client.Destroy

procedure TPLTCP_Client.Send(aMSG: String);
begin
  Socket.SendBlock(aMSG);
end; // TPLTCP_Client.Send

procedure TPLTCP_Client.StartSocket(IP: String; Port: Integer);
begin
  Socket:= TTCPBlockSocket.Create;
  Socket.Connect(IP,InttoStr(Port));
  Start;
end; // TPLTCP_Client.Start

procedure TPLTCP_Client.Execute;
var
  aMessage:String;
begin
  while not Terminated do begin

    aMessage := Socket.RecvBlock(1000);
    if Socket.LastError = 0 then begin
      LastMessage:=aMessage;
//      writeln(LastMessage);
      @DoResivedMsg;
      //Synchronize(@DoResivedMsg);
    end;
  end;
end; // TPLTCP_Client.Execute

procedure TPLTCP_Server.Execute;
var
  newconnection:TSocket;
  newclient:TPLTCP_Client;
begin
  while not Terminated do begin
    newconnection := Socket.Accept;
    if (not Terminated) then begin
      newclient:=ClientList.AddClient('','');
      newclient.Socket := TTCPBlockSocket.Create;
      newclient.Socket.Socket:= newconnection;
      newclient.ID:='Test1';
      newclient.FromHost:=newclient.Socket.GetLocalSinIP;
      newclient.Start;

      if Assigned(OnAddClient) then OnAddClient(newclient);
    end
    else
      break;
  end;
end; // TPLTCP_Server.Execute

{ TPLTCP_Server }
constructor TPLTCP_Server.Create(CreateSuspended: Boolean; const StackSize: SizeUInt);
begin
  inherited Create(CreateSuspended,StackSize);
  ClientList:=TPLTCP_ClientList.Create;
  ClientList.Server:=self;
  fOnAddClient:=nil;
  fOnReseived:=nil;
  fPort:=-1;
  fHost:='';
end; // TPLTCP_Server.Create

destructor TPLTCP_Server.Destroy;
begin
  ClientList.Free;
  inherited Destroy;
end; // TPLTCP_Server.Destroy

procedure TPLTCP_Server.StartServer(const aHost:string; const aPort:Integer);
begin
  Host:=aHost;
  Port:=aPort;
  Socket := TTCPBlockSocket.Create;
  Socket.Bind(aHost,InttoStr(aPort));
  Socket.Listen;
  Start;
end; // TPLTCP_Server.StartServer

procedure TPLTCP_Server.Stop;
begin
  Terminate;
end; // TPLTCP_Server.Stop

end.

